################################################################################
# MIT License
#
# Copyright (c) 2016 Meezio SAS <dev@meez.io>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
################################################################################

from werkzeug.exceptions import default_exceptions
from werkzeug.exceptions import HTTPException
from flask import Flask
from flask import request
from flask import jsonify
from .config import config


class API(Flask):
    def __init__(self, import_name):
        super(API, self).__init__(import_name)
        self.prefix = ""

        # Format internal error message to JSON
        for code in default_exceptions.keys():
            # self.error_handler_spec[None][code] = self.__make_json_error
            self.register_error_handler(code, self.__make_json_error)

        # Set headers for all response
        self.after_request(self.__setHeaders)

    # Add prefix to route
    def route(self, rule, **options):
        def decorator(f):
            f.__name__ = self.prefix + "_" + f.__name__
            endpoint = options.pop('endpoint', None)
            self.add_url_rule(self.prefix + rule, endpoint, f, **options)
            return f
        return decorator

    # Return error information if debug_level = 1,
    # if 0 return HTTP code 500 and generic HTTP message
    def __make_json_error(self, ex):
        if isinstance(ex, HTTPException):
            if hasattr(ex, 'verbose') and ex.verbose:
                response = jsonify(message=ex.description)
            else:
                response = jsonify(message=str(ex))

            response.status_code = ex.code
        else:
            if config.debug:
                response = jsonify(message=str(ex))
            else:
                response = jsonify(message="Internal Server Error")
            response.status_code = 500

        return response

    # Function to be run after each request to set headers
    def __setHeaders(self, response):
        response.headers.add('Server', config.server_name)
        response.headers.add('Access-Control-Allow-Origin', '*')
        if request.method == 'OPTIONS':
            response.headers.add('Access-Control-Allow-Methods', 'DELETE, GET, HEAD, PATCH, POST, PUT')
            response.headers["Access-Control-Allow-Headers"] = "Authorization, Content-Type"
        return response

api = API(__name__)



################################################################################
# MIT License
#
# Copyright (c) 2016 Meezio SAS <dev@meez.io>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
################################################################################

import json
import datetime
import time
from flask import request
from jose import jwt
from jose.exceptions import ExpiredSignatureError, JWTError
from urllib.request import urlopen
from .logger import Logger
from .exceptions import Http401Exception, Http403Exception
from .config import config
from .cache import cache


def Authorization():
    """ Check that JSON Web Token (JWT) passed through ``Authorization`` header or through  query parameter 'token' is valid.
    The JWT MUST be provided by an OpenID Connect provider and be passed as a Bearer token :

    .. code:: bash

        Authorization: Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6IkpvaG4gRG9lIiwiYWRtaW4iOnRydWV9.TJVA95OrM7E2cBab30RMHrHDcEfxjoYZgeFONFh7HgQ


    To validate signature, the publics keys are retrieved by fetching the issuer URL at ``/.well-known/openid-configuration`` and are store in cache for further use.

    :return: The claims contained in the JWT body.
    :rtype: dict
    :raises apicore.Http401Exception: If they is no Authorization header.
    :raises apicore.Http403Exception: If Authorization header is not valid.
    """

    token = request.args.get('token')
    if not token:
        # Authorization required
        if "Authorization" not in request.headers:
            raise Http401Exception("Neither Authorization header nor 'token' query parameter are present")

        # Get JWT from HTTP headers
        auth = request.headers.get("Authorization").split()
        if len(auth) is not 2 and auth[0] is not "Bearer":
            raise Http403Exception("Malformed Authorization HTTP Header : format MUST match Bearer TOKEN.")

        token = auth[1]

    # Find the token issuer
    try:
        tmp = jwt.get_unverified_claims(token)
        kid = jwt.get_unverified_header(token).get('kid')
        issuer = tmp.get('iss')
    except JWTError as ex:
        raise Http403Exception(str(ex))
    if not issuer:
        raise Http403Exception("'iss' claim missing in JWT.")

    # If whitelist is empty every issuers are allowed except those from blacklist
    if config.issWhitelist:
        if issuer not in config.issWhitelist:
            raise Http403Exception("Issuer '{}' not in Whitelist.".format(issuer))
    if config.issBlacklist:
        if issuer in config.issBlacklist:
            raise Http403Exception("Issuer '{}' is blacklisted.".format(issuer))

    # Get the issuer's keys
    key = cache.get("{}@{}".format(issuer, kid))
    if not key:
        Logger.info("Caching keys for issuer '{}'".format(issuer))
        oidcConf = getJSON("{}/.well-known/openid-configuration".format(issuer))
        keys = getJSON(oidcConf.get("jwks_uri")).get("keys")
        # Cache expires in 365 or less day depending on 'exp' attribute in JWS
        ts = time.mktime(datetime.datetime.utcnow().timetuple()) + (365 * 24 * 60 * 60)  # timespamp now + 365days
        for k in keys:
            expire = k.get('exp', ts)
            kkid = k.get('kid')

            # convert timespam from ms to s
            if expire > 4000000000:
                expire = expire // 1000

            cache.set("{}@{}".format(issuer, kkid), k, expire)
            key = cache.get("{}@{}".format(issuer, kid))

    try:
        userProfile = jwt.decode(token, key, options={"verify_aud": False, "verify_iss": False, "verify_sub": False, "verify_exp": config.tokenExpire})
        return userProfile
    except ExpiredSignatureError as ex:
        raise Http403Exception(str(ex), True)
    except JWTError as ex:
        raise Http403Exception(str(ex))


def getJSON(url):
    data = urlopen(url).read().decode("utf-8")
    return json.loads(data)



################################################################################
# MIT License
#
# Copyright (c) 2016 Meezio SAS <dev@meez.io>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
################################################################################

import pickle
import redis
from .logger import Logger
from .config import config


class Cache:
    """ Cache for the application.
    To use by importing the instance :

    exemple:
        .. code::

            from apicore import cache

            key = "my_data"
            date = {"color": "orange", "flag": True}
            cache.set(key, data)
            print(cache.get(key))


    .. note::

        If ``redis`` URI is configured the cache is store in redis server, otherwise it is cache in memory.

    """
    def __init__(self):
        self.module = None

    def set(self, key, value, expire=None):
        """

        :param str key: the key referencing the data
        :param value: the data to store in cache
        :param integer expire: Expire at a given timestamp in seconde.
        """
        self._getModule().set(key, value, expire)

    def get(self, key):
        """

        :param str key: the key referencing the data
        """
        return self._getModule().get(key)

    def _getModule(self):
        if not self.module:
            if config.redis:
                self.module = _redis()
            else:
                self.module = _memory()

        return self.module


class _redis:
    def __init__(self):
        Logger.info("Create Redis connection pools for cache : '{}'".format(config.redis))
        self.conn = redis.StrictRedis.from_url(config.redis)

    # expire: timespam in seconde
    def set(self, key, value, expire):
        self.conn.set(key, pickle.dumps(value))
        if expire:
            self.conn.expireat(key, expire)

    def get(self, key):
        data = self.conn.get(key)

        if data:
            return pickle.loads(data)
        else:
            return None


class _memory:
    def __init__(self):
        Logger.info("Using memory for cache")
        self.data = dict()

    def set(self, key, value, expire):
        self.data[key] = value

    def get(self, key):
        return self.data.get(key)


cache = Cache()



################################################################################
# MIT License
#
# Copyright (c) 2016 Meezio SAS <dev@meez.io>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
################################################################################

import yaml
from .logger import Logger
from .exceptions import Http500Exception


class Config:
    """ Manage configuration values.
    To use by importing the instance :

    exemple:
        .. code::

            from apicore import config

            print(config.server_name)


    """
    def __init__(self):
        self.data = dict()
        self.__setDefault()

    def load(self, confFile='config.yaml'):
        """ Load config file from filesystem.

        :param str string: Path to config file.
        """
        try:
            with open(confFile) as f:
                data = yaml.load(f)
                if data:
                    self.data.update(data)
        except FileNotFoundError:
                Logger.error("Configuration file not found")

    def __getattr__(self, name):
        if name in self.data:
            return self.data[name]
        else:
            raise Http500Exception("Config value '{}' does not exist.".format(name))

    def __setDefault(self):
        # Service name
        self.data['server_name'] = "API server"
        # Active debug
        self.data['debug'] = True
        # Accept all JWT from issuers from this list
        self.data['issWhitelist'] = None
        # Reject all JWT from issuers from this list
        self.data['issBlacklist'] = None
        # Validate token against Expire date/time
        self.data['tokenExpire'] = True
        # redis://:password@localhost:6379/10
        self.data['redis'] = None

config = Config()



################################################################################
# MIT License
#
# Copyright (c) 2016 Meezio SAS <dev@meez.io>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
################################################################################

from werkzeug import exceptions
from .logger import Logger

__all__ = ['Http400Exception', 'Http401Exception', 'Http403Exception', 'Http404Exception', 'Http409Exception', 'Http500Exception']

"""
Exceptions are provided to send regular HTTP error code and messages.

To raise this kind of exception: ::

    raise Core400Exception("Additionals informations")  # Return "Bad Request" in HTTP Response and print "Additionals informations" to log
    raise Core404Exception("Additionals informations", True)  # Return "Additionals informations" in HTTP Response

To handle this kind of excpetion: ::

    try:
        ...
    except Core404Exception as ex:
        print ex.code  # HTTP return code
        print ex.name  # HTTP message
        print ex.description  # Additionals informations

"""


class Http400Exception(exceptions.BadRequest):
    """ Create a 400 *Bad Request* exception.

    :param str infos: A message describing the error.
    :param boolean verbose: True to send infos in HTTP response.
    """
    def __init__(self, description=None, verbose=False):
        self.verbose = verbose
        exceptions.BadRequest.__init__(self, description, None)
        Logger.info("HTTP {} : {}".format(400, self.description))


class Http401Exception(exceptions.Unauthorized):
    """ Create a 401 *Unauthorized* exception.

    :param str infos: A message describing the error.
    :param boolean verbose: True to send infos in HTTP response.
    """
    def __init__(self, description=None, verbose=False):
        self.verbose = verbose
        exceptions.Unauthorized.__init__(self, description, None)
        Logger.info("HTTP {} : {}".format(401, self.description))


class Http403Exception(exceptions.Forbidden):
    """ Create a 403 *Forbidden* exception.

    :param str infos: A message describing the error.
    :param boolean verbose: True to send infos in HTTP response.
    """
    def __init__(self, description=None, verbose=False):
        self.verbose = verbose
        exceptions.Forbidden.__init__(self, description, None)
        Logger.info("HTTP {} : {}".format(403, self.description))


class Http404Exception(exceptions.NotFound):
    """ Create a 404 *Not Found* exception.

    :param str infos: A message describing the error.
    :param boolean verbose: True to send infos in HTTP response.
    """
    def __init__(self, description=None, verbose=False):
        self.verbose = verbose
        exceptions.NotFound.__init__(self, description, None)
        Logger.info("HTTP {} : {}".format(404, self.description))


class Http409Exception(exceptions.Conflict):
    """ Create a 409 *Conflict* exception.

    :param str infos: A message describing the error.
    :param boolean verbose: True to send infos in HTTP response.
    """
    def __init__(self, description=None, verbose=False):
        self.verbose = verbose
        exceptions.Conflict.__init__(self, description, None)
        Logger.info("HTTP {} : {}".format(409, self.description))


class Http500Exception(exceptions.InternalServerError):
    """ Create a 500 *Internal Server Error* exception.

    :param str infos: A message describing the error.
    :param boolean verbose: True to send infos in HTTP response.
    """
    def __init__(self, description=None, verbose=False):
        self.verbose = verbose
        exceptions.InternalServerError.__init__(self, description, None)
        Logger.error(self.description)



"""
apicore
=======

Set of core libraries usefull for building REST API and Microservices based on Flask.

The code is open source, release under MIT and written in Python 3.

.. code:: bash

    pip install apicore


Features
--------

* Cross-origin resource sharing (CORS) ready
* Data caching with redis server or direct in memory
* Configuration file loader
* A simple Logger
* Raise exception conform to HTTP status codes
* Authorization using JSON Web Token(JWT) issued from an OpenID Connect provider.

Links
-----

* `documentation <http://apicore.readthedocs.io/en/latest/>`_
* `sources <https://github.com/meezio/apicore>`_
* `compagny <https://meez.io>`_
"""

from setuptools import setup

setup(
    name='apicore',
    version='1.0.0',
    packages=['apicore'],
    author="Meezio SAS",
    author_email="dev@meez.io",
    description="Core lib for REST API",
    long_description=__doc__,
    url='https://github.com/meezio/apicore',
    classifiers=[
        "Intended Audience :: Developers",
        "Development Status :: 5 - Production/Stable",
        "Environment :: No Input/Output (Daemon)",
        "Framework :: Flask",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: French",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Server",
        "Topic :: Software Development :: Libraries"
    ],
    license="MIT",
    install_requires=[
        'termcolor',
        'Flask',
        'PyYAML',
        'python-jose',
        'redis'
    ]
)
